# Plugin for TWiki Enterprise Collaboration Platform, http://TWiki.org/
#
# Copyright (C) 2015-2018 Peter Thoeny, peter[at]thoeny.org 
# Copyright (C) 2015-2018 TWiki Contributors. All Rights Reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version. For
# more details read LICENSE in the root of this distribution.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details, published at
# http://www.gnu.org/copyleft/gpl.html
#
# As per the GPL, removal of this notice is prohibited.

package TWiki::Plugins::LoadAppPlugin::Core;

# =========================
sub new {
    my ( $class, $debug ) = @_;

    my $this = {
          Debug  => $debug,
          Map    => _mapFromCfg( $TWiki::cfg{Plugins}{LoadAppPlugin}{Map} ),
        };
    bless( $this, $class );
    $this->_writeDebug( "new: constructor" );

    return $this;
}

# =========================
sub VarLOADAPP
{
    my ( $this, $session, $params, $theTopic, $theWeb ) = @_;
    my $default = $params->{_DEFAULT};
    my $action  = $params->{action} || 'load';
    $this->_writeDebug( "VarLOADAPP($action)" );
    my $text = '';
    if( $action eq 'load' ) {
        $text = $this->_loadAppTopic( $session, $default, $params, $theTopic, $theWeb );
    } elsif( $action ) {
        $text = "LOADAPP ERROR: Unrecogized action parameter $action";
    } else {
        $text = '';
    }
    if( $this->{Debug} ) {
        $text =~ /^(.{0,80})/s;
        $this->_writeDebug( "VarLOADAPP($default, $action) return:\n$1..." );
    }
    return $text;
}

# =========================
sub _loadAppTopic
{
    my ( $this, $session, $id, $params, $theTopic, $theWeb ) = @_;

    $this->_writeDebug( "_loadTopic($id)" );

    # Map from ID to app topic
    unless( $id ) {
        return "LOADAPP ERROR: Required ID is missing";
    }
    my $appWeb = '';
    my $appTopic = $this->{Map}{$id};
    unless( $appTopic ) {
        return "LOADAPP ERROR: ID $id not found";
    }
    ( $appWeb, $appTopic ) = TWiki::Func::normalizeWebTopicName( $appWeb, $appTopic );
    unless( TWiki::Func::topicExists( $appWeb, $appTopic ) ) {
        return "LOADAPP ERROR: ID $id not found";
    }

    # SMELL: Do INCLUDE using undocumented system variables, needed to circumvent access control.
    # However, circumvent access control is OK because we limit the topics by the ID to topic
    # mapping. Below code is largely duplicating from TWiki::INCLUDE.
    my %saveTags = %{$session->{SESSION_TAGS}};
    my $prefsMark = $session->{prefs}->mark();
    $session->{SESSION_TAGS}{INCLUDINGWEB} = $appWeb;
    $session->{SESSION_TAGS}{INCLUDINGTOPIC} = $appTopic;

    # Read topic text
    my ( $meta, $text ) = TWiki::Func::readTopic( $appWeb, $appTopic );
    $text =~ s/^[\r\n]+/\n/;
    $text =~ s/[\r\n]+$//;

    # Handle sections
    my $section = $params->{section};
    if( !$section ) {
       $text =~ s/.*?%STARTINCLUDE%//s;
       $text =~ s/%STOPINCLUDE%.*//s;
    }
    my( $ntext, $sections ) = _parseSections( $text );
    my $interesting = ( defined $section );
    if( $interesting || scalar( @$sections )) {
        # Rebuild the text from the interesting sections
        $text = '';
        foreach my $s ( @$sections ) {
            if( $section && $s->{type} eq 'section' && $s->{name} eq $section) {
                $text .= substr( $ntext, $s->{start}, $s->{end}-$s->{start} );
                $disableFixLinks = 1 if( $s->{disablefixlinks} );
                $interesting = 1;
                last;
            } elsif( $s->{type} eq 'include' && !$section ) {
                $text .= substr( $ntext, $s->{start}, $s->{end}-$s->{start} );
                $interesting = 1;
            }
        }
    }
    # If there were no interesting sections, restore the whole text
    $text = $ntext unless $interesting;

    # Handle parameters and default variables
    my $tagsWithDefault = undef;
    $tagNameRegex = TWiki::Func::getRegularExpression( 'tagNameRegex' );
    $text =~ s/(%)($tagNameRegex)({ *default=")(.*?[^\\]?)(" *})(%)/
               $tagsWithDefault->{$2} = _unescapeQuotes( $4 );
               "$1$2$6"/ge;
    foreach my $k ( keys %$params ) {
        next if( $k eq '_RAW' );
        # copy params into session tags
        $session->{SESSION_TAGS}{$k} = $params->{$k};
        # remove captured tag with default
        delete $tagsWithDefault->{$k};
    }
    foreach my $k ( keys %$tagsWithDefault ) {
        # copy left over captured tags with default into session tags
        $session->{SESSION_TAGS}{$k} = $tagsWithDefault->{$k};
    }

    # Handle common variables in context of current web
    $text = TWiki::Func::expandCommonVariables( $text, $theTopic, $theWeb, $meta );

    # Restore session variables
    %{$session->{SESSION_TAGS}} = %saveTags;
    $session->{prefs}->restore( $prefsMark );

    return $text;
}

# =========================
sub _unescapeQuotes {
    my $text = shift;
    $text =~ s/\\(["'])/$1/g;
    return $text;
}

# =========================
# This is copied from TWiki.pm
sub _parseSections {
    #my( $text _ = @_;
    my %sections;
    my @list = ();

    my $seq = 0;
    my $ntext = '';
    my $offset = 0;
    foreach my $bit (split(/(%(?:START|END)SECTION(?:{.*?})?%)/, $_[0] )) {
        if( $bit =~ /^%STARTSECTION(?:{(.*)})?%$/) {
            require TWiki::Attrs;
            my $attrs = new TWiki::Attrs( $1 );
            $attrs->{type} ||= 'section';
            $attrs->{name} = $attrs->{_DEFAULT} || $attrs->{name} || '_SECTION'.$seq++;
            delete $attrs->{_DEFAULT};
            my $id = $attrs->{type}.':'.$attrs->{name};
            if( $sections{$id} ) {
                # error, this named section already defined, ignore
                next;
            }
            # close open unnamed sections of the same type
            foreach my $s ( @list ) {
                if( $s->{end} < 0 && $s->{type} eq $attrs->{type} &&
                      $s->{name} =~ /^_SECTION\d+$/ ) {
                    $s->{end} = $offset;
                }
            }
            $attrs->{start} = $offset;
            $attrs->{end} = -1; # open section
            $sections{$id} = $attrs;
            push( @list, $attrs );
        } elsif( $bit =~ /^%ENDSECTION(?:{(.*)})?%$/ ) {
            require TWiki::Attrs;
            my $attrs = new TWiki::Attrs( $1 );
            $attrs->{type} ||= 'section';
            $attrs->{name} = $attrs->{_DEFAULT} || $attrs->{name} || '';
            delete $attrs->{_DEFAULT};
            unless( $attrs->{name} ) {
                # find the last open unnamed section of this type
                foreach my $s ( reverse @list ) {
                    if( $s->{end} == -1 &&
                          $s->{type} eq $attrs->{type} &&
                         $s->{name} =~ /^_SECTION\d+$/ ) {
                        $attrs->{name} = $s->{name};
                        last;
                    }
                }
                # ignore it if no matching START found
                next unless $attrs->{name};
            }
            my $id = $attrs->{type}.':'.$attrs->{name};
            if( !$sections{$id} || $sections{$id}->{end} >= 0 ) {
                # error, no such open section, ignore
                next;
            }
            $sections{$id}->{end} = $offset;
        } else {
            $ntext .= $bit;
            $offset = length( $ntext );
        }
    }

    # close open sections
    foreach my $s ( @list ) {
        $s->{end} = $offset if $s->{end} < 0;
    }

    return( $ntext, \@list );
}

# =========================
sub _writeDebug
{
    my ( $this, $msg ) = @_;
    return unless( $this->{Debug} );
    TWiki::Func::writeDebug( "- LoadAppPlugin::Core::$msg" );
}

# =========================
sub _mapFromCfg {
    my $cfg = shift;
    my $map;
    foreach my $k ( keys %$cfg ) {
        my $text = $cfg->{$k};
        foreach $item ( split( /,\s*/, $text ) ) {
            if( $item =~ /^([^:]+):\s*(.*)$/ ) {
                $map->{$1} = $2;
            }
        }
    }
    return $map;
}

# =========================
1;
